#include "ACLfipa.hxx"
#include "parser.h"
#include "ACLsub.hxx"
#include <string>

int ACLfipa::type()
{
  if(root() != NULL)
    return root()->key();
  else
    return 0;
}

ACLnode * ACLfipa::sender()
{
  ACLnode **list;
  list = find(root(),Parser::SENDER);
  if(*list == NULL)
    return NULL;
  else
    return (*list)->left();
}

ACLnode * ACLfipa::receiver()
{
  ACLnode **list;
  list = find(root(),Parser::RECEIVER);
  if(*list == NULL)
    return NULL;
  else
    return (*list)->left();
}

ACLnode * ACLfipa::receiver(int i)
{
  ACLnode *t;
  t = receiver();

  if(t == NULL)
    return NULL;

  if(t->key() == Parser::AGENT){
    if(i == 1)
      return t;
    else
      return NULL;
  }

  if(t->key() == Parser::SET)
    return set(i,t);

  return NULL;
}

int ACLfipa::sender_id()
{
  if(sender() != NULL)
    return sender()->left()->value();
  return 0;
}

int ACLfipa::receiver_id(int i)
{
  if(receiver(i) != NULL)
    return receiver(i)->left()->value();
  return 0;
}


int ACLfipa::receiver_num()
{
  ACLnode *t;
  t = root()->left();

  while(t != NULL){
    if( t->key() == Parser::RECEIVER ){
      switch(t->left()->key()){
      case Parser::SET:
	return set_num(t->left());
      case Parser::AGENT:
	return 1;
      default:
	return 0;
      }
    }
    t = t->right();
  }
    return 0;
}

ACLnode * ACLfipa::content()
{
  ACLnode **list;
  list = find(root(),Parser::CONTENT);
  if(*list == NULL)
    return NULL;
  else
    return (*list)->left();
}

int ACLfipa::set_num(ACLnode *n)
{
  int i;
  ACLnode *t;

  if(n == NULL)
    return 0;

  if(n->key() != Parser::SET)
    return 0;

  i = 0;
  t = n->left();
  while(t != NULL){
    t = t->right();
    i++;
  }

  return i;
}

ACLnode * ACLfipa::set(int i,ACLnode *n)
{
  int j;
  ACLnode *t;

  if(n == NULL || i > set_num(n))
    return NULL;

  if(n->key() != Parser::SET)
    return NULL;

  if(set_num(n) <= 0)
    return NULL;

  t = n->left();
  for(j=1;j<i;j++)
    t = t->right();
  return t;
}

ACLnode * ACLfipa::find_s(int k)
{
  ACLnode **list;
  list = ACLtree::find(root(),k);
  if(*list == NULL)
    return NULL;
  else
    return (*list);
}

void ACLfipa::print_tree()
{
  if(root() == NULL)
    return;

  int j;

  printf("m-type: %s\n",change(root()));

  if( sender() != NULL)
    printf("sender: %s\n",change(sender()->left()));

  if( receiver() != NULL){
    printf("reciver_num:%d\n", receiver_num());
    printf("reciver: ");
    for(j=1;j <= receiver_num();j++){
      printf("%s ",change(receiver(j)->left()));
    }
    printf("\n");
  }
  if( content() != NULL){
    printf("content:\n");
    ACLtree::print_tree_sub(0,content());
  }

  printf("\n");
}

void ACLfipa::out_sub(char *out,ACLnode *n)
{
  char *buff = new char[100];
  ACLnode *t;
  t = n;

  if(t->key() != Parser::INTEGER &&
     t->key() != Parser::TINTEGER)
    strcpy(out," (");
  else
    strcpy(out," ");

  strcat(out,change(t));

  if(t->left() != NULL){
    strcpy(out," (");
    strcat(out,change(t));
    out_sub(buff,t->left());
    strcat(out,buff);
  }

  if(t->key() != Parser::INTEGER &&
     t->key() != Parser::TINTEGER)
    strcat(out,")");

  if(t->right() != NULL){
    out_sub(buff,t->right());
    strcat(out,buff);
  }  

  delete [] buff;
}

void ACLfipa::out(char *message)
{
  if(root() == NULL)
    return;

  char *s = new char[50];
  char *r = new char[50];
  char *c = new char[300];
  char *t = new char[300];

  *s = '\0';
  if(sender() != NULL){
    strcpy(s,":sender");
    out_sub(t,sender());
    strcat(s,t);
  }

  *r = '\0';
  if(receiver() != NULL){
    strcpy(r,":receiver");
    out_sub(t,receiver());
    strcat(r,t);
  }

  *c = '\0';
  if(content() != NULL){
    strcpy(c,":content (");
    out_sub(t,content());
    strcat(c,t);
    strcat(c,")");
  }

  sprintf(message,"(%s %s %s %s)  ",change(root()),s,r,c);
  
  delete [] s;
  delete [] r;
  delete [] c;
  delete [] t;
}

 
