// main.cxx
#include "common.hxx"
#include "basic.hxx"
#include "Config.hxx"
#include "Address.hxx"
#include "Input.hxx"
#include "Output.hxx"
#include "LongUDPSocket.hxx"
#include "ObjectPool.hxx"
#include "objects.hxx"
#include "sim_class.h"
using namespace Rescue;

std::ofstream flogout; // MA(00/06/02)
std::ofstream ferrout; // MA(00/07/22)

int traffic_main(int argc, char** argv);
void initRealMap(ObjectPool& pool, RealMap& real_map);
void commands(S32& time, Input& input, Output& output, LongUDPSocket& socket, Address& to, ObjectPool& pool, RealMap& real_map);
void commands(S32& time, Header type, Input& input, ObjectPool& pool, RealMap& real_map);
void setMotionOnRoad(MovingObject* agent, Input& input, ObjectPool& pool, RealMap& real_map);
void setMotionOnNode(MovingObject* agent, Input& input, ObjectPool& pool, RealMap& real_map);
void setMotionOnBuilding(S32& time, MovingObject* agent, Input& input, ObjectPool& pool, RealMap& real_map);
void setStay(MovingObject* agent, RealMap& real_map);
void reviseRealMap(RealMap& real_map);

int main(int argc, char** argv)
{

#ifdef TRAFFIC_LOG
	// open log file
	flogout.open("traffic.log");
	flogout << "# traffic log\n" << std::flush;
#endif
	ferrout.open("error.log");

	try {
		int result = traffic_main(argc, argv);

#ifdef TRAFFIC_LOG
		// close log file
		flogout.close();
#endif
		ferrout.close();

		return result;
	} catch(std::exception& e) {
		fprintf(stderr, "%s\n", (const char*)e.what());
	} catch(...) {
		fprintf(stderr, "error...in main\n");
	}

#ifdef TRAFFIC_LOG
	// close log file
	flogout.close();
#endif
	ferrout.close();

	return 1;
}

int traffic_main(int argc, char** argv)
{
	const char* host = "localhost";
	Config config;

	// analysis of arguments
	int i;
	int count = 0;
	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(strcmp(argv[i], "-file") == 0) {
				if(++i < argc) {
					config.readConfigFile(argv[i]);
				} else {
					fprintf(stderr, "error: '-file'\n");
				}
			} else {
				if(++i < argc) {
					if(!config.setValue(argv[i-1]+1, argv[i]))
						fprintf(stderr, "error: '%s'\n", argv[i]);
				} else {
					fprintf(stderr, "error: '%s'\n", argv[i-1]);
				}
			}
		} else {
			switch(count++) {
			case 0:
				host = argv[i];
				break;
			default:
				fprintf(stderr, "error: '%s'\n", argv[i]);
				break;
			}
		}
	}

	// socket creation
	LongUDPSocket socket(config.textlogname_samplesimulator());
	Input input;
	Output output;
	output.setWait(config.send_udp_wait());
	Address to(host, config.port());	// 

	// ³å
	printf("connecting...\n");
	output.clear();	// ϥХåեˤ
	output.put(SK_CONNECT);	// ϥХåե SK_CONNECT (4ХȤ)Ϥ
	Output::Cursor base = output.put(~(S32)0);	// body ΥȤƤȤꤢߡͤϤƤơbase ˤΰ֤򵭲롣
	output.put(0);	// ϥХåե 0 Ϥ
	output.setSize(base);	// base ǵ֤ˡbase 餳ޤǤ put Хȿꤹ롣
	output.put(HEADER_NULL);
	socket.send(output, to, -1); // output  put Х to 롣

	// å롼
	S32 time = 0;
	ObjectPool pool;

	// ߥ졼Υ֥
	RealMap real_map;

	for(;;) {
		try {
			socket.receive(input);	// ǡinput ߤ롣
			for(;;) {
				Header header = input.get();	
							// input (4ХȤ)ǡФheader 
				if(header == HEADER_NULL)
					break;

				S32 size = input.get();
				Input::Cursor start = input.cursor();	// ɤ߽ФƤ֤򵭲

				switch(header) {
				default:
					printf("Unknown header.\n");
					break;
				case KS_CONNECT_ERROR:
					printf("Connection failed.\n");
					return 1;
				case KS_CONNECT_OK:
					printf("Connect Ok.\n");
					// ack ֤
					output.clear();
					output.put(SK_ACKNOWLEDGE);
					output.put(~(S32)0);
					output.put(HEADER_NULL);
					socket.send(output, to, -1);

					to = socket.addressRecievedFrom();	// ʹߤѹ
					pool.restructure(0, input);			// ɥǥ롣
					initRealMap(pool, real_map);   // ߥ졼⥪֥Ȥν
					break;
				case KS_COMMANDS:
					
					try {
					commands(time, input, output, socket, to, pool, real_map);
					} catch(std::exception& e) {
						fprintf(stderr, "Exception: %s\n", (const char*)e.what());
					} catch(...) {fprintf(stderr, "Exception in commands\n"); break;}
#ifdef TRAFFIC_LOG
					flogout << std::flush;
#endif
					break;
				case KS_UPDATE:
					time = input.get();

#ifdef TRAFFIC_LOG
					flogout << "# Received KS_UPDATE(t=" << time << ")\n" << std::flush;
#endif

					try {
					pool.restructure(time, input);
					} catch(...) {fprintf(stderr, "Exception in pool restructure\n"); throw;}

					printf("time = %ld ... KS_UPDATE\n", (long)time);

					break;
				}

				input.setCursor(start);
				input.skip(size);						// size Хȥå
			}
		} catch(std::exception& e) {
			fprintf(stderr, "%s\n", (const char*)e.what());
		} catch(...) {
			fprintf(stderr, "Unknown exception\n");
		}
	}
	
	return 0;
}

// ߥ졼⥪֥Ȥν
void initRealMap(ObjectPool& pool, RealMap& real_map)
{
  Objects::const_iterator p;
  SimCars::iterator p_scar;
  SimNodes::iterator p_snode;
  SimRoads::iterator p_sroad;

  // poolΥݥ󥿤򥪥֥˳꿶äƥԡ
  for(p = pool.objects().begin(); p != pool.objects().end(); p++){

    // 
    MovingObject* p_mo = dynamic_cast<MovingObject*>(*p);
    if(p_mo != 0){
		real_map.sim_car.insert(std::make_pair(p_mo->id(),SimCar(p_mo)));
      continue;
    }
    // Ρ
    Node* p_node = dynamic_cast<Node*>(*p);
    if(p_node != 0){
		real_map.sim_node.insert(std::make_pair(p_node->id(),SimNode(p_node)));
      continue;
    }
    // 
    Road* p_road = dynamic_cast<Road*>(*p);
    if(p_road != 0){
		real_map.sim_road.insert(std::make_pair(p_road->id(),SimRoad(p_road)));
      continue;
    }
  }

  // ֺǽν
  for(p_snode = real_map.sim_node.begin(); p_snode != real_map.sim_node.end(); p_snode++){
    printf("node: %ld\n",p_snode->first);
    p_snode->second.setEdges();
    p_snode->second.setAccessKind();
    p_snode->second.setShortcutToTurn();
    p_snode->second.setPocketToTurnAcross();
    p_snode->second.setSignalTiming();
  }
  for(p_sroad = real_map.sim_road.begin(); p_sroad != real_map.sim_road.end(); p_sroad++){
    printf("road: %ld\n",p_sroad->first);
    p_sroad->second.setHead();
    p_sroad->second.setTail();
    p_sroad->second.setHeadConnectNum(real_map.sim_node);
    p_sroad->second.setTailConnectNum(real_map.sim_node);
  }
  for(p_scar = real_map.sim_car.begin(); p_scar != real_map.sim_car.end(); p_scar++){
    printf("car: %ld\n",p_scar->first);
  }
}

void commands(S32& time, Input& input, Output& output, LongUDPSocket& socket, Address& to, ObjectPool& pool, RealMap& real_map)
{
    // 괺ΥȤpositionpositionExtra򥳥ԡ
    SimCars::iterator p;
	for(p = real_map.sim_car.begin(); p != real_map.sim_car.end(); p++){
	  p->second.setPosition();
	  p->second.setPositionExtra();
	  if(p->second.positionExtra<0){
	    printf("posExtra: %ld\n",p->second.positionExtra);
	  }
	  p->second.motion = 0;
	  if(POSITION_HISTORY) p->second.clearPositionHistory();
	}

	// KS_COMMANDS 
	S32 newTime = input.get();
	if(newTime < time) {
		printf("warning: new_time=%ld, old_time=%ld\n", (long)newTime, (long)time);
		return;
	}
	time = newTime;

	printf("time = %ld ... KS_COMMANDS\n", (long)time);
#ifdef TRAFFIC_LOG
	flogout << "t=" << time << "  ...received KS_COMMANDS\n";
#endif

	Header type;
	int check_flag = 0;
	while(type = input.get(), type != HEADER_NULL) {
		check_flag = 1;
		S32 size = input.get();
		Input::Cursor start = input.cursor();
		commands(time, type, input, pool, real_map);
		input.setCursor(start);
		input.skip(size);
	}
	if(check_flag == 0) return;

	// ߥ졼ν

/* DEBUG
	for(p = real_map.sim_car.begin(); p != real_map.sim_car.end(); p++){
		if((p->second.position >= 822) && (p->second.motion == 1))
			ferrout << "CONTRADICTION: CAR=" << p->first << ",POS=" << p->second.position 
			        << "\n"<< std::flush;
    }
*/

	reviseRealMap(real_map);

#if defined(DEBUG) || defined(TRAFFIC_LOG)
	// ưֵϿ
	int csize = real_map.sim_car.size();
	std::map<Id, Id> carpos;
	std::map<Id, S32> carposExtra;
	for(p = real_map.sim_car.begin(); p != real_map.sim_car.end(); p++){
		carpos.insert(std::make_pair(p->first, p->second.position));
		carposExtra.insert(std::make_pair(p->first, p->second.positionExtra));
	}
#endif	

	// 1ʬ֤θήߥ졼
#ifdef DEBUG
	printf("1 minute simulation start\n");
#endif

	real_map.traffic();

#ifdef DEBUG
	printf("1 minute simulation end\n");
#endif

	for(p = real_map.sim_car.begin(); p != real_map.sim_car.end(); p++){
	  if(p->second.motion != 0){
	    p->second.revisePosition(time,pool);
	    p->second.revisePositionExtra(time);
		if(POSITION_HISTORY) p->second.recordPositionHistory(time);

/* TEST OF POSITION HISTORY
Objects history = (p->second).positionHistory;
Objects::iterator p = history.begin();
while(p != history.end()) {
	ferrout << (*p)->id() << " ";
	p++;
}
ferrout << "\n" << std::flush;
*/

	  }
	}

#if defined(DEBUG) || defined(TRAFFIC_LOG)
	// ưΰ
	int moveCount = 0;
	for(p = real_map.sim_car.begin(); p != real_map.sim_car.end(); p++){
		std::map<Id, Id>::iterator it_c = carpos.find(p->first);
		std::map<Id, Id>::iterator it_ce = carposExtra.find(p->first);
		if(it_c->second != p->second.position || 
			it_ce->second != p->second.positionExtra) {
			moveCount++;
		} else {
#ifdef DEBUG
			printf(" stop_at=%d ", p->second.position);
#endif
#ifdef TRAFFIC_LOG
			if(p->second.motion == 1) {
				SimRoads::iterator now = real_map.sim_road.find(p->second.position);
				flogout << "movable car(" << p->first << ") stop at=" << p->second.position;
				int pe;
				if(p->second.direct == HEAD) {
					pe = p->second.positionExtra;
					flogout << "(H:" << pe << "/" << now->second.pointer->length() << ")";
				} else { 
					pe = now->second.pointer->length() - p->second.positionExtra;
					flogout << "(T:" << pe << "/" << now->second.pointer->length() << ")";
				}
				if(p->second.aim_num < p->second.step+2) {
					flogout << " [why don't go to goal node?]\n";
				} else {
					SimRoads::iterator next = real_map.sim_road.find(p->second.aim[p->second.step+2]);
					if(next == real_map.sim_road.end())
							flogout << " [why don't go to building?]\n";
					else {
						int al;
						if(p->second.aim[p->second.step+1]==next->second.head)
							al = next->second.linesAliveTail;
						else al = next->second.linesAliveHead;
						if(al == 0) {
							flogout << " [because alive lines of next road(" << next->first << ")=" << al << " :";
							flogout << "width=" << next->second.pointer->width()
								    << ",block=" << next->second.pointer->block() << "]\n";
						} else {
							int f = p->second.front_id;
							if(f == NP) {
								int space;
								if((p->second).isPedestrian())
										space = real_map.pjudgeInroad(p->first);
								else space = real_map.judgeInroad(p->first);
								if(space == 0) {
									flogout << " [alive lines of next roads(" << next->first << ")=" << al << " but there is no space]\n";
								} else {
									flogout << " [*why don't go to alive lines of next road(" << next->first << ")=" << al << " with space?:";
									flogout << "front_car=no";
									if(p->second.pocket == 1) flogout << ",pocket = in pocket";
									else flogout << ",pocket=out of pocket";
									int ac = real_map.getAccess(now->first, next->first);
									switch(ac) {
									case RIGHT: flogout << ",go to right"; break;
									case LEFT: flogout << ",go to left"; break;
									case STRAIGHT: flogout << ",go to straight"; break;
									default: flogout << ", direction?"; break;
									}
									int oncoming = real_map.judgeOncomingCar(p->first);
									if(oncoming == 1) flogout << ",exist oncoming car";
									
									flogout << "]\n";
								}
							} else {
								int dist = p->second.between;
								if(dist <= DIST_BET_CARS*1000) {
									flogout << " [because front car is too near:";
									flogout << "front_car=" << f;
									flogout << ",dist=" << dist;
									if(p->second.pocket == 1) flogout << ",pocket = in pocket";
									else flogout << ",pocket=out of pocket";
									flogout << "]\n";
								} else { 
									flogout << " [*why don't go in spite of enough front space:";
									flogout << "front_car=" << f;
									flogout << ",dist=" << dist;
									if(p->second.pocket == 1) flogout << ",pocket = in pocket";
									else flogout << ",pocket=out of pocket";
									flogout << "]\n";
								}
							}
						}
					}
				}
			}
#endif
		}
	}
#endif

#ifdef DEBUG
	printf("car moved=%d\n", moveCount);
#endif
#ifdef TRAFFIC_LOG
	flogout << "car moved=" << moveCount << "\n";
#endif

	// ߥ졼̤
	output.clear();
	output.put(SK_UPDATE);
	Output::Cursor base = output.put(~(S32)0);
	pool.output(output, time);
	output.setSize(base);
	output.put(HEADER_NULL);
	socket.send(output, to, time);
}

void commands(S32& time, Header type, Input& input, ObjectPool& pool, RealMap& real_map)
{
	Id sender;
	while(sender = input.get(), sender != 0) {
		S32 size = input.get();
		Input::Cursor start = input.cursor();

		Object* o = pool.get(sender);	// ɥǥ뤫顢sender  ID ˻ĥ֥ȤĥФ
		MovingObject* agent = dynamic_cast<MovingObject*>(o);
		Road* p_road;
		Node* p_node;
		Building* p_build;
		if(agent != 0) {
			switch(type) {
			case AK_MOVE:
			        o = pool.get(agent->position()->id());
			        p_road = dynamic_cast<Road*>(o);
				if(p_road != 0){
				  setMotionOnRoad(agent, input, pool, real_map);
				  break;
				}
				p_node = dynamic_cast<Node*>(o);
				if(p_node != 0){
				  setMotionOnNode(agent, input, pool, real_map);
				  //setStay(agent,real_map);
				  break;
				}
				p_build = dynamic_cast<Building*>(o);
				if(p_build != 0){
				  setMotionOnBuilding(time, agent, input, pool, real_map);
				  break;
				}
				else setStay(agent,real_map);
				break;
			default:
				setStay(agent, real_map);
			}
		}
		
		input.setCursor(start);
		input.skip(size);
	}
}

// Ȥηϩץ(ϤRoadξ) ... (00/12)
void setMotionOnRoad(MovingObject* agent, Input& input, ObjectPool& pool, RealMap& real_map)
{
   SimCars::iterator p = real_map.sim_car.find(agent->id());
   p->second.motion = 1;                                         // ưե饰ON
   p->second.aim.clear();                                        // ŪϤ򥯥ꥢ

   int length = 0, overlap = 0; // v0.22
   p->second.aim.push_back(p->second.position);                  // ֤߰aim[0]
   length++; // v0.22

#ifdef DEBUG
   printf("R: %d | ", p->second.position);
#endif

   Id aim_id;
   Object *o = pool.get(p->second.position);
   Road *o_road = dynamic_cast<Road*>(o);

   while(1) {
      aim_id = input.get();
#ifdef DEBUG
	  printf("%d ", aim_id);
#endif

      if(aim_id == 0) break;
      if((aim_id == o_road->head()->id()) ||
         (aim_id == o_road->tail()->id())) {
         p->second.aim.push_back(aim_id);
		 length++; // v0.22

#ifdef DEBUG
		 printf("$ ");
#endif
         break;
      }
   }
   if(aim_id != 0) {
      while(aim_id = input.get()) {
         p->second.aim.push_back(aim_id);
		 length++; // v0.22

#ifdef DEBUG
         printf("%d ", aim_id);
#endif	 
         if(aim_id == p->second.position) { // v0.22
			 overlap = length;
#ifdef DEBUG
		 printf("# ");
#endif
		 }
      }
   }

   if(overlap > 0) { // v0.22
	   std::vector<Id>::iterator it = p->second.aim.begin();
	   p->second.aim.erase(it+1, it+overlap);
   }
   
   p->second.aim_num = p->second.aim.size();
#ifdef DEBUG
   printf(" (%d)\n", p->second.aim_num); 
#endif
   if(p->second.aim_num < 2) {
	   p->second.motion = -2;
#ifdef DEBUG
	   printf("aim sequence is short(R):%d\n", p->second.aim_num);
#endif
   }
#ifdef DEBUG
   if(p->second.aim_num == 2) printf("aim sequence is 2(R)\n");
#endif
}

// Ȥηϩץ(ϤNodeξ)...(00/12)
void setMotionOnNode(MovingObject* agent, Input& input, ObjectPool& pool, RealMap& real_map)
{
   SimCars::iterator p = real_map.sim_car.find(agent->id());
   p->second.motion = 1;                                         // ưե饰ON
   p->second.aim.clear();                                        // ŪϤ򥯥ꥢ  

   Id aim_id;
#ifdef DEBUG
   printf("N: ");
#endif
   while(1) {
      aim_id = input.get();
#ifdef DEBUG
      printf("%d ", aim_id);
#endif
      if(aim_id == 0) break;
      Object *o = pool.get(aim_id);

	  // (00/12)
	  Building *b = dynamic_cast<Building*>(o);
	  if(b != 0) {
//		  printf("N-->B\n");
		  p->second.position = aim_id;
		  p->second.motion = 2;
		  return;
	  }

      Road *o_road = dynamic_cast<Road*>(o);
      if(o_road == 0) continue;
      if(o_road->head()->id() == agent->position()->id()) {
         p->second.position = aim_id;
         p->second.aim.push_back(aim_id);
         p->second.positionExtra = 0;
#ifdef DEBUG
         printf("$ ");
#endif
         break;   
      } else if(o_road->tail()->id() == agent->position()->id()) {
         p->second.position = aim_id;
         p->second.aim.push_back(aim_id);
         p->second.positionExtra = o_road->length();
#ifdef DEBUG
         printf("$ ");
#endif
         break;
      }
   } 
   if(aim_id != 0) {
      while(aim_id = input.get()) {
         p->second.aim.push_back(aim_id);
#ifdef DEBUG
         printf("%d ", aim_id);
#endif
      }
   }

   p->second.aim_num = p->second.aim.size();
#ifdef DEBUG
   printf(" (%d)\n", p->second.aim_num);
#endif
   if(p->second.aim_num < 2) {
	   p->second.motion = -1;
#ifdef DEBUG
	   printf("aim sequence is short(N):%d\n", p->second.aim_num);
#endif
   }
#ifdef DEBUG
   if(p->second.aim_num == 2) printf("aim sequence is 2(N)\n");
#endif
}

void setMotionOnBuilding(S32& time, MovingObject* agent, Input& input, ObjectPool& pool, RealMap& real_map)
{
	SimCars::iterator p = real_map.sim_car.find(agent->id());
	p->second.aim.clear();
	p->second.motion = 0;
	
	if(((p->second).isCivilian()) &&
		(((Civilian*)(p->second.pointer))->buriedness() > 0)) {
		p->second.position = agent->position()->id();
		p->second.motion = -3;
	} else {
		Id aim_id = input.get();
		if(aim_id == 0) {
//			printf("B--B\n");
			p->second.aim_num = 0;
			p->second.motion = 2;
			return;
		}
		Object *o = pool.get(aim_id);
		Node *n = dynamic_cast<Node*>(o);
		if(n != 0) {
//			printf("B-->N\n");
			p->second.position = aim_id;
			p->second.positionExtra = 0;
			p->second.aim_num = 0;
			p->second.revisePosition(time,pool);
			p->second.revisePositionExtra(time);
			setMotionOnNode(agent, input, pool, real_map);
		} else {
			p->second.position = agent->position()->id();
			p->second.motion = -4;
		}
	}
}

// ưʤȤϰưե饰motionOFFˤ
void setStay(MovingObject* agent, RealMap& real_map)
{
  SimCars::iterator p = real_map.sim_car.find(agent->id());
  p->second.motion = 0;
}

////////////////////////////////////////////////////////////////////////////////////////////////