// sim_class.h
#ifndef SIMCLASS
#define SIMCLASS

#include <iostream> // MA(00/06/02)
#include <fstream>	// MA(00/06/02)

// debug and log ... MA(00/06/03)
//#define TRAFFIC_LOG
//#define DEBUG

///////////////////////////////

namespace Rescue{

///////////////////////////////

#define MAX_NUM_OF_EDGES	6
#define MAX_NUM_OF_LINES	6

#define MAX_NUM_OF_PLINES	1

#define RIGHT		0
#define LEFT		1
#define STRAIGHT	2
#define UNDEFINED	-1
#define ILLEGAL		100

#define HEAD 0
#define TAIL 1

#define DIST_BET_CARS 4
#define MAX_SPEED 16
#define ACCELERATION 3

#define DIST_BET_PEDESTRIANS 1
#define PMAX_SPEED 1.2
#define PACCELERATION 0.4

const Id NP=1000000;

#define POSITION_HISTORY 1

struct SimCar{
  MovingObject* pointer;     //
  Id position;               //
  S32 positionExtra;         //
  int flag;                  //
  int direct;                //
  S32 speed;                 //
  S32 between;               //
  Id front_id;               //
  Id back_id;                //
  int step;                  // the number of aims achieved in a minute
  int pocket;                //
  int line;                  //
  std::vector<Id> aim;       //
  int aim_num;               //
  int motion;                //
  Objects positionHistory;

public:
  SimCar(MovingObject* p_mo){
    pointer = p_mo;
    position = 0;
    positionExtra = 0;
    flag = 0;
    direct = 0;
    speed = 0;
    between = 0;
    front_id = 0;
    back_id = 0;
    step = 0;
    pocket = 0;
    line = 0;
    aim_num = 0;
    motion = 0; 
  }

  void setPosition(){ position = pointer->position()->id(); }
  void setPositionExtra(){ positionExtra = pointer->positionExtra(); }
  void revisePosition(S32& time, ObjectPool& pool){ pointer->setPosition(time,pool.get(position)); }
  void revisePositionExtra(S32& time){ pointer->setPositionExtra(time,positionExtra); }

  bool isCivilian() { if(pointer->type() == TYPE_CIVILIAN) return true; else return false; }

  bool isPedestrian() { 
	  if((pointer->type() == TYPE_CIVILIAN) &&
		 ((pointer->id()<2327)||(pointer->id()>2336))) return true;
	  else return false;
  }

  void addPositionHistory(Node * n) { positionHistory.push_back(n); }
  void clearPositionHistory() { positionHistory.clear(); }
  void recordPositionHistory(S32 time) { pointer->setPositionHistory(time, positionHistory); }
};

typedef std::map<Id,SimCar> SimCars;

///////////////////////////////

struct SimNode{
  Node* pointer;             // 
  Id edges[MAX_NUM_OF_EDGES];// road IDs connected to a road
  int connect_road_num;      // the number of roads connected to a node
  int accessKind[MAX_NUM_OF_EDGES][MAX_NUM_OF_EDGES];      // Ⱦ
  std::vector<int> shortcutToTurn; // number vector of shortcuts for left turn 
  std::vector<int> pocketToTurnAcross;
  std::vector<int> signalTiming;
  
public:
  SimNode(Node* p_node){
    pointer = p_node;
    connect_road_num = 0;
  }

  void setEdges(){
    int count = 0;
    Objects::const_iterator p;
    // m_edges֥ȤΤɤΤߤФIDֹedges˳Ǽ
    for(p = pointer->edges().begin(); p != pointer->edges().end(); p++){
      Road* p_road = dynamic_cast<Road*>(*p);
      if(p_road != 0) edges[count++] = p_road->id();
    }
    connect_road_num = count;
  }

  void setAccessKind(){	// This is a temporal function must be modified in future
    int i;
    switch(connect_road_num){
    case 1:
      accessKind[0][0] = UNDEFINED;
      break;
    case 2:
		accessKind[0][0] = accessKind[1][1] = UNDEFINED;
      accessKind[0][1] = accessKind[1][0] =  STRAIGHT;
      break;
    case 3:
      accessKind[0][0] = accessKind[1][1] = accessKind[2][2] = UNDEFINED;
      accessKind[1][0] = accessKind[2][1] = RIGHT;
      accessKind[0][1] = accessKind[1][2] = LEFT;
      accessKind[0][2] = accessKind[2][0] = STRAIGHT;
      break;
    case 4:
      accessKind[0][0] = accessKind[1][1] = accessKind[2][2] = accessKind[3][3] = UNDEFINED;
      accessKind[0][3] = accessKind[1][0] = accessKind[2][1] = accessKind[3][2] = RIGHT;
      accessKind[0][1] = accessKind[1][2] = accessKind[2][3] = accessKind[3][0] = LEFT;
      accessKind[0][2] = accessKind[1][3] = accessKind[2][0] = accessKind[3][1] = STRAIGHT;
      break;
    case 5:
      for(i=0;i<5;i++)  accessKind[i][i] =UNDEFINED;
      accessKind[0][3] = accessKind[0][4] = accessKind[1][0] = accessKind[1][4] = accessKind[2][1] = accessKind[3][2] = accessKind[4][3] = RIGHT;
      accessKind[0][1] = accessKind[1][2] = accessKind[2][3] = accessKind[2][4] = accessKind[3][0] = accessKind[3][4] = accessKind[4][0] = LEFT;
      accessKind[0][2] = accessKind[1][3] = accessKind[2][0] = accessKind[3][1] = accessKind[4][1] = accessKind[4][2] = STRAIGHT;
      break;
    case 6:
      for(i=0;i<6;i++) accessKind[i][i] = UNDEFINED;
      accessKind[0][4] = accessKind[0][5] = accessKind[1][0] = accessKind[1][5] = accessKind[2][0] = accessKind[2][1] = accessKind[3][1] = accessKind[3][2] = accessKind[4][2] = accessKind[4][3] = accessKind[5][3] = accessKind[5][4] = RIGHT;
      accessKind[0][1] = accessKind[0][2] = accessKind[1][2] = accessKind[1][3] = accessKind[2][3] = accessKind[2][4] = accessKind[3][4] = accessKind[3][5] = accessKind[4][0] = accessKind[4][5] = accessKind[5][0] = accessKind[5][1] = LEFT;
      accessKind[0][3] = accessKind[1][4] = accessKind[2][5] = accessKind[3][0] = accessKind[4][1] = accessKind[5][2] = STRAIGHT;
      break;
    default:
      printf("connect_riad_num is too large.\n");
      break;
    }
  }

  void setShortcutToTurn(){
    Bytes::const_iterator p;
    for(p = pointer->shortcutToTurn().begin(); p != pointer->shortcutToTurn().end(); p++){
      shortcutToTurn.push_back((int)*p);
    }
  }

  void setPocketToTurnAcross(){
    Bytes::const_iterator p;
    for(p = pointer->pocketToTurnAcross().begin(); p != pointer->pocketToTurnAcross().end(); p++){
      pocketToTurnAcross.push_back((int)*p);
    }
  }

  void setSignalTiming(){
    Bytes::const_iterator p;
    for(p = pointer->signalTiming().begin(); p != pointer->signalTiming().end(); p++){
      signalTiming.push_back((int)*p);
    }
  }
};

typedef std::map<Id,SimNode> SimNodes;

///////////////////////////////

struct SimRoad{
  Road* pointer;             // 
  Id head;                   // 
  Id tail;                   // 
  S32 linesAliveHead;        // 
  S32 linesAliveTail;        // 
  Id head_pocket;            // last car ID in right turn pocket (head direction)
  Id tail_pocket;            //                                  (tail direction)
  Id head_last_id[MAX_NUM_OF_LINES];        // last car ID for each line (head direction)
  Id tail_last_id[MAX_NUM_OF_LINES];        //                           (tail direction)
  int head_connect_num;      // 
  int tail_connect_num;      // 
  Id phead_last_id[MAX_NUM_OF_PLINES];	// last pedestrian ID for each line (head direction) 
  Id ptail_last_id[MAX_NUM_OF_PLINES];	// last pedestrian ID for each line (tail direction) 

public:
  SimRoad(Road* p_road){
    pointer = p_road;
    head = 0;
    tail = 0;
    linesAliveHead = 0;
    linesAliveTail = 0;
    head_pocket = 0;
    tail_pocket = 0;
    head_connect_num = 0;
    tail_connect_num = 0;
  }

  void setHead(){ head = pointer->head()->id(); }
  void setTail(){ tail = pointer->tail()->id(); }

  void setLinesAlive(){
	/*
	 * same processing for pointer->medianStrip() == 0 and 1
	 */
	double lineWidth = ((double)pointer->width())/
								 (pointer->linesToHead()+pointer->linesToTail());
	double blockWidth = ((double)pointer->block())/2;

	double linesBlockedRate = blockWidth/lineWidth;

	S32 linesBlocked = (S32)floor(linesBlockedRate+0.5);

	linesAliveHead = pointer->linesToHead()-linesBlocked;
	if(linesAliveHead < 0) linesAliveHead = 0;
	linesAliveTail = pointer->linesToTail()-linesBlocked;
	if(linesAliveTail < 0) linesAliveTail = 0;
#ifdef DEBUG
	if(pointer->block() > 0)
		printf("block/width=%ld/%ld\n", pointer->block(), pointer->width());
#endif
  }

  void setHeadConnectNum(SimNodes& sim_node){
    SimNodes::iterator p = sim_node.find(head);
    int i;
    for(i = 0; i < p->second.connect_road_num; i++){
      if(p->second.edges[i] == pointer->id()){
	head_connect_num = i;
	break;
      }
    }
    if(i == p->second.connect_road_num){
      printf("SimRoad::head_connect_num error\n");
      exit(1);
    }
  }

  void setTailConnectNum(SimNodes& sim_node){
    SimNodes::iterator p = sim_node.find(tail);
    int i;
    for(i = 0; i < p->second.connect_road_num; i++){
      if(p->second.edges[i] == pointer->id()){
	tail_connect_num = i;
	break;
      }
    }
    if(i == p->second.connect_road_num){
      printf("SimRoad::tail_connect_num error\n");
      exit(1);
    }
  }
};

typedef std::map<Id,SimRoad> SimRoads;

//////////////////////////////////////

struct RealMap{
  SimCars sim_car;
  SimNodes sim_node;
  SimRoads sim_road;
public:
  void InitEnv();
  void InitCar();
  void InitLines();
  void InitFront();
  void InitRoad();
  void traffic();
  // move for car
  void move(Id);
  void move_to_pocket(Id);
  void move_in_pocket(Id);
  void move_normal(Id);
  void move_front(Id);
  void move_at_cross(Id);
  void move_signal(Id);
  void move_on_blue(Id);
  void move_node_across(Id);
  void move_no_signal(Id);
  void move_node(Id);
  void move_inroad(Id);
  void move_toward_building(Id, Id); // MA
  // move for pedestrian
  void pmove(Id);
  void pmove_normal(Id);
  void pmove_front(Id);
  void pmove_at_cross(Id);
  void pmove_no_signal(Id);
  void pmove_node(Id);
  void pmove_inroad(Id);
  void pmove_toward_building(Id, Id);
  void pmove_signal(Id);
  // aux
  int getAccess(Id,Id);
  int judgeOncomingCar(Id);
  int getAliveLines(Id,Id,Id);
  int pass(Id); // not used
  // aux for car
  int *getLineInroad(Id);
  int judgeInroad(Id);
  // aux for pdestrian
  int *pgetLineInroad(Id);
  int pjudgeInroad(Id);
};

}
#endif
