// simworld_setup.cxx
#include "common.hxx"
#include "basic.hxx"
#include "Config.hxx"
#include "Address.hxx"
#include "Input.hxx"
#include "Output.hxx"
#include "LongUDPSocket.hxx"
#include "ObjectPool.hxx"
#include "objects.hxx"
#include "sim_class.h"
using namespace Rescue;

extern std::ofstream flogout; // MA(00/06/02)
extern std::ofstream ferrout;

void reviseRealMap(RealMap& real_map) // ʬϤ˸ƤФ
{
  SimRoads::iterator p_road;
  for(p_road = real_map.sim_road.begin(); p_road != real_map.sim_road.end(); p_road++){
    p_road->second.setLinesAlive();         // linesAliveHeadlinesAliveTail
  }

  real_map.InitEnv();
}

///////////////////////////////////////////////////////////////////////////////
// Ķ
void RealMap::InitEnv()
{
	InitCar();
	InitLines();
	try {
	InitFront();
	} catch(...) { fprintf(stderr, "EXCEPTION in INITFRONT\n"); throw; }
	InitRoad();
}

///////////////////////////////////////////////////////////////////////////////
// ֤ν

void RealMap::InitCar() // MAʬϤ˸ƤФ
{
	SimCars::iterator p;
	SimRoads::iterator now;

#ifdef TRAFFIC_LOG
	int cnum = 0, movable = 0, unmovable = 0, blockade = 0; 
#endif

	for(p=sim_car.begin();p!=sim_car.end();p++){

#ifdef TRAFFIC_LOG
		cnum++;
#endif

		// ˤưǽ֤Ƚ(000717)
		if(p->second.motion == 1) {
			now = sim_road.find(p->second.position);
			S32 alive;
			if(p->second.direct == HEAD) alive = now->second.linesAliveHead;
			else alive = now->second.linesAliveTail;
			if(alive == 0) p->second.motion = -3;
#ifdef DEBUG
		printf("[%ld]", alive);
#endif
		}
		// END

		if(p->second.motion!=1){
			p->second.speed = 0;
			if(p->second.motion == -1) {
#ifdef DEBUG
				printf("N"); 
#endif
#ifdef TRAFFIC_LOG
				unmovable++;
#endif
				p->second.motion = 0; 
			} else if(p->second.motion == -2) {
#ifdef DEBUG
				printf("R"); 
#endif
#ifdef TRAFFIC_LOG
				unmovable++;
#endif
				p->second.motion = 0; 
			} else if(p->second.motion == -3) {
#ifdef DEBUG
				printf("B"); 
#endif
#ifdef TRAFFIC_LOG
				unmovable++; blockade++;
#endif
				p->second.motion = 0;
			} else if(p->second.motion == -4) {
#ifdef DEBUG
				printf("X"); 
#endif
#ifdef TRAFFIC_LOG
				unmovable++; 
#endif
				p->second.motion = 0;
			} else if(p->second.motion == 2) {
#ifdef DEBUG
				printf("G"); 
#endif
#ifdef TRAFFIC_LOG
				movable++; 
#endif
			}
#ifdef DEBUG
			else printf("#");				// MA(00/06/01)
#endif
		}else{
			p->second.flag = 0;
			p->second.between = 0;
			p->second.step = 0;
			p->second.pocket = 0;
			
			if(now->second.head==p->second.aim[1]) {
				p->second.direct = HEAD; 
#ifdef DEBUG
				printf("0");
#endif
#ifdef TRAFFIC_LOG
				movable++;
#endif
			}
			else if(now->second.tail==p->second.aim[1]) {
				p->second.direct = TAIL; 
#ifdef DEBUG
				printf("1");
#endif
#ifdef TRAFFIC_LOG
				movable++;
#endif
			}
		}
	}
#ifdef DEBUG
	printf("\n");
#endif
#ifdef TRAFFIC_LOG
	flogout << "car=" << cnum
	      << ",active=" << movable+unmovable
		  << ",movable=" << movable 
		  << ",unmovable=" << unmovable
		  << " (in_collapse=" << blockade << ")\n";
	double trouble;
	if(movable+unmovable == 0) trouble = 0;
	else trouble = (double)unmovable/(movable+unmovable);
	flogout << "rate of unmovable=" << trouble << "\n";
#endif
}

///////////////////////////////////////////////////////////////////////////////
// line
void RealMap::InitLines() // øƤФ->ʬϤѹ
{
  int access,lines,alive,con;
  SimCars::iterator p;
  SimRoads::iterator now,next;
  SimNodes::iterator node_id;

  for(p=sim_car.begin();p!=sim_car.end();p++){
    if(p->second.motion==1){
      // begin pedestrian
		if((p->second).isPedestrian()) {
			p->second.line = 0;
			continue;
		}
	  // end
      now = sim_road.find(p->second.position);
	  // 000717
	  if(p->second.aim_num > p->second.step + 2) {
		next = sim_road.find(p->second.aim[(p->second.step)+2]);
		if(next==sim_road.end()) access = STRAIGHT; // ʪξľ
		else access = getAccess(now->first,next->first);
	  } else access = STRAIGHT;
	  if(access != RIGHT && access != LEFT && access != STRAIGHT) {
		printf("Warning: route plan is invalid.\n");
		ferrout << "Warning in InitLines: invalid route plan for the car " 
			<< p->first << " at " << now->first << "\n" << std::flush;
		access = STRAIGHT; // 
		p->second.motion = 0; continue; // EMERGENCY
//		exit(1);
	  }
	  // 
      if(p->second.direct==HEAD){
		node_id=sim_node.find(now->second.head);
		con = now->second.head_connect_num;
		alive = now->second.linesAliveHead;  // >0
		lines = (now->second.pointer)->linesToHead();
		if(access==RIGHT){
			if(node_id->second.pocketToTurnAcross.size()!=0){
				if(node_id->second.pocketToTurnAcross[con*2]>=1 && 
					p->second.positionExtra<node_id->second.pocketToTurnAcross[con*2+1]){
					p->second.pocket = 1;
				}else{
					p->second.line=lines-1;
				}
			}else{
				p->second.line=lines-1;
			}
		} else if(access == LEFT) {
			p->second.line = lines-alive;
		} else { // access == STRAIGHT
			p->second.line = (lines-alive)+(int)(p->first%alive);
		}
      }else{
		node_id=sim_node.find(now->second.tail);
		con = now->second.tail_connect_num;
		alive = now->second.linesAliveTail; // >0
		lines = (now->second.pointer)->linesToTail();
		if(access==RIGHT){
			if(node_id->second.pocketToTurnAcross.size()!=0){
				if(node_id->second.pocketToTurnAcross[con*2]>=1 && 
					(now->second.pointer)->length() - p->second.positionExtra
					<node_id->second.pocketToTurnAcross[(con*2+1)]){
					p->second.pocket = 1;
				}else{
					p->second.line=lines-1;
				}
			}else{
				p->second.line=lines-1;
			}
		} else if(access == LEFT) {
			p->second.line = lines-alive;
		} else { // access == STRAIGHT
			p->second.line = (lines-alive)+(int)(p->first%alive);
		}
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
// ֤ȸ֡μִ֤׻
void RealMap::InitFront() // øƤФ->ʬϤѹ
{
  int max_back,max_front;
  SimCars::iterator p,front;
  SimRoads::iterator now;

  for(p=sim_car.begin();p!=sim_car.end();p++){
	if(p->second.motion==1){
	  TypeId type1 = (p->second.pointer)->type();
      now = sim_road.find(p->second.position);
//	  try {

      max_front=(now->second.pointer)->length();
      max_back=(now->second.pointer)->length();
/*
	  } catch(...) { 
		  fprintf(stderr, "EXCEPTION in MAX of INITFRONT:");
		  ferrout << "EXCEPTION in MAX of INITFRONT:"<< std::flush;
		  fprintf(stderr, "C=%d, P=%d, NOW=%d, %d, %d\n", p->first, p->second.position, now->first, 
			      now->second.head_connect_num, max_front);
		  ferrout << "CAR=" << p->first << ",POS=" << p->second.position << ",NOW=" << now->first
			      << "," << now->second.head_connect_num << "," << max_front << "\n"<< std::flush;
		  throw; 
	  }
*/
      p->second.front_id = NP;
      p->second.back_id = NP;
      if(p->second.direct==HEAD) p->second.between = p->second.positionExtra;
      else{
		p->second.between=(now->second.pointer)->length() - p->second.positionExtra;
      }

      for(front=sim_car.begin();front!=sim_car.end();front++){
		//
	    TypeId type2 = (front->second.pointer)->type();
		if(type1 == TYPE_CIVILIAN) {
			if((p->second).isPedestrian()) {
				if(type2 != TYPE_CIVILIAN) continue;
				else if((type2 == TYPE_CIVILIAN) && !((front->second).isPedestrian())) continue;
			} else {
				if((front->second).isPedestrian()) continue;
			}
		} else {
			if((front->second).isPedestrian()) continue;
		}
	    //
		if(front->first!=p->first && 
			front->second.motion==1 && 
			front->second.position==p->second.position && 
			front->second.direct==p->second.direct &&
			((p->second.pocket==0 && front->second.pocket==0 && 
			  p->second.line==front->second.line) ||
			  (p->second.pocket==1 && front->second.pocket==1))) {
			if(p->second.direct==HEAD){
				if(p->second.positionExtra > front->second.positionExtra &&
					p->second.positionExtra-front->second.positionExtra<=max_front){
					p->second.front_id = front->first; 
					max_front = p->second.positionExtra-front->second.positionExtra;
					p->second.between = max_front;
				}else if(p->second.positionExtra < front->second.positionExtra && 
						front->second.positionExtra-p->second.positionExtra
						<=max_back){
					p->second.back_id = front->first;
					max_back=front->second.positionExtra-p->second.positionExtra;
				}
			}else{
				if(p->second.positionExtra < front->second.positionExtra &&
					front->second.positionExtra-p->second.positionExtra<=max_front){
					p->second.front_id = front->first;
					max_front=front->second.positionExtra-p->second.positionExtra;
					p->second.between = max_front;
				}else if(p->second.positionExtra > front->second.positionExtra &&
						p->second.positionExtra-front->second.positionExtra<=max_back){
					p->second.back_id = front->first;
					max_back = p->second.positionExtra-front->second.positionExtra;
				}
			}

		}

      }

    }
  }
}

///////////////////////////////////////////////////////////////////////////////
// ɽ:ƥ饤ˤǸID, ޥݥåȤˤǸID
void RealMap::InitRoad() // øƤФ->ʬϤѹ
{
  int i,max_head[MAX_NUM_OF_LINES],max_head_pocket,max_tail[MAX_NUM_OF_LINES],max_tail_pocket,lines;
  int max_phead[MAX_NUM_OF_PLINES], max_ptail[MAX_NUM_OF_PLINES];
  SimCars::iterator p;
  SimRoads::iterator now;
  for(now=sim_road.begin();now!=sim_road.end();now++){
    for(i=0;i<MAX_NUM_OF_LINES;i++){
		now->second.head_last_id[i] = NP;
		now->second.tail_last_id[i] = NP; 
		max_head[i] = 0;
		max_tail[i] = (now->second.pointer)->length();
    }
	for(i=0;i<MAX_NUM_OF_PLINES;i++){
		now->second.phead_last_id[i] = NP;
		now->second.ptail_last_id[i] = NP; 
		max_phead[i] = 0;
		max_ptail[i] = (now->second.pointer)->length();
    }
    now->second.head_pocket = NP;
    now->second.tail_pocket = NP;
    max_head_pocket = 0;
    max_tail_pocket = (now->second.pointer)->length();
    for(p=sim_car.begin();p!=sim_car.end();p++){
      if(p->second.motion==1 && p->second.position==now->first){
		  if((p->second).isPedestrian()) {
			if(p->second.direct==HEAD){
				if(p->second.positionExtra>=max_phead[0]){
					now->second.phead_last_id[0] = p->first;
					max_phead[0] = p->second.positionExtra;
				}
			} else {
				if(p->second.positionExtra<=max_ptail[0]){
					now->second.ptail_last_id[0] = p->first;
					max_ptail[0] = p->second.positionExtra;
				}
			}
		  } else {
			if(p->second.direct==HEAD){
				if(p->second.pocket==1 && p->second.positionExtra>=max_head_pocket){
					now->second.head_pocket = p->first;
					max_head_pocket = p->second.positionExtra;
				}else{
					lines = (now->second.pointer)->linesToHead();
					for(i=0;i<lines;i++){
						if(p->second.line==i && p->second.positionExtra>=max_head[i]){
							now->second.head_last_id[i] = p->first;
							max_head[i] = p->second.positionExtra;
						}
					}
				}
			}else{
				if(p->second.pocket==1 && p->second.positionExtra<=max_tail_pocket){
					now->second.tail_pocket = p->first;
					max_tail_pocket = p->second.positionExtra;
				}else{
					lines = (now->second.pointer)->linesToTail();
					for(i=0;i<lines;i++){
						if(p->second.line==i && p->second.positionExtra<=max_tail[i]){
							now->second.tail_last_id[i] = p->first;
							max_tail[i] = p->second.positionExtra;
						}
					}
				}
			}
		  }
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
